;.pagewidth  120t
;*********************************************************************
;* Title:  HCS12_Serial_Monitor.asm        Copyright (c) Motorola 2003
;*********************************************************************
;* Author: Jim Sibigtroth - Motorola TSPG - 8/16 Bit Division
;* Author: Jim Williams - Motorola TSPG - 8/16 Bit Division
;*
;* Description: Bootloader/Monitor program for HCS9S12
;* bootloader will reside in 2K of block protected memory at the
;* end of the memory map of an HCS9S12 MCU
;*
;* Since this code is located in the vector space, all interrupt
;* vectors will be mirrored to the pseudo vector table in user
;* erasable and reprogrammable flash memory just before the start
;* of the protected bootloader code.
;*
;* If a non-FFFF user reset vector is programmed into the
;* pseudo-reset vector, the bootloader will jump to that routine
;* so the user can control all options including write-once bits.
;*
;* There are 22 primitive monitor commands to allow a debug program
;* in a host PC to work through the SCI instead of through a BDM pod.
;*
;* Documentation: HCS9S12/D Preliminary Data Sheet and  HCS12
;* Preliminary Family Reference Manual Volume 1 (HCS12RM1/D)
;*
;*
;* Revision History: not yet released
;* Rev #     Date      Who     Comments
;* -----  -----------  ------  ---------------------------------------
;*  1.0    14-APR-03   JPW     Port from SerialMon9S08
;*  1.1    15-APR-03   JPW     modified device command to send ID in hex
;*  1.2    17-APR-03   JPW     Removed unused ascii routines
;*  1.3    18-APR-03   JPW     Fixed device ID command
softwareID1:  equ   $0418     ;Software revision (date)
softwareID2:  equ   $2003     ;Software revision (year)
;*
;*
;*********************************************************************
;*********************************************************************
;* Motorola reserves the right to make changes without further notice
;* to any product herein to improve reliability, function, or design.
;* Motorola does not assume any liability arising out of the
;* application or use of any product, circuit, or software described
;* herein; neither does it convey any license under its patent rights
;* nor the rights of others.  Motorola products are not designed,
;* intended, or authorized for use as components in systems intended
;* for surgical implant into the body, or other applications intended
;* to support life, or for any other application in which the failure
;* of the Motorola product could create a situation where personal
;* injury or death may occur.  Should Buyer purchase or use Motorola
;* products for any such intended or unauthorized application, Buyer
;* shall indemnify and hold Motorola and its officers, employees,
;* subsidiaries, affiliates, and distributors harmless against all
;* claims, costs, damages, and expenses, and reasonable attorney fees
;* arising out of, directly or indirectly, any claim of personal
;* injury or death associated with such unintended or unauthorized
;* use, even if such claim alleges that Motorola was negligent
;* regarding the design or manufacture of the part.
;*
;* Motorola is a registered trademark of Motorola, Inc.
;*********************************************************************

            XDEF Startup         ;make this symbol visible to the linker
;*********************************************************************
;* Include standard definitions that are common to all derivatives
;*********************************************************************
             base    10           ;ensure default number base to decimal
             nolist               ;turn off listing
             include "HCS12_Serial_Monitor.inc"
             list                 ;turn listing back on
;*********************************************************************
;* One device specific NVM definition below should be uncommented
;*********************************************************************
Dx512:
;eeprom_start:  equ   $0400         ;eeprom start
;eeprom_end:    equ   $07FF         ;eeprom end
;sector_size:   equ   $0400         ;Flash sector size
;Flash_blks:    equ   $04           ;# of flash blks
;lowest_page:   equ   $20           ;lowest page
;Pages_Blk:     equ   $08           ;banks per block
Dx256:
;eeprom_start:  equ   $0400         ;eeprom start
;eeprom_end:    equ   $0FFF         ;eeprom end
;sector_size:   equ   $0200         ;Flash sector size
;Flash_blks:    equ   $04           ;# of flash blks
;lowest_page:   equ   $0030         ;lowest page
;Pages_Blk:     equ   $04           ;pages per block
Dx128:
;eeprom_start:  equ   $0000         ;eeprom start (no EEprom)
;eeprom_end:    equ   $0000         ;eeprom end  (no EEprom)
;sector_size:   equ   $0400         ;Flash sector size
;Flash_blks:    equ   $01           ;# of flash blks
;lowest_page:   equ   $38           ;lowest page
;Pages_Blk:     equ   $08           ;pages per block
E128:
;eeprom_start:  equ   $0000         ;eeprom start (no EEprom)
;eeprom_end:    equ   $0000         ;eeprom end  (no EEprom)
;sector_size:   equ   $0400         ;Flash sector size
;Flash_blks:    equ   $01           ;# of flash blks
;lowest_page:   equ   $38           ;lowest page
;Pages_Blk:     equ   $08           ;pages per block
Dx64:
;eeprom_start:  equ   $0C00         ;eeprom start
;eeprom_end:    equ   $0FFF         ;eeprom end
;sector_size:   equ   $0200         ;Flash sector size
;Flash_blks:    equ   $01           ;# of flash blks
;lowest_page:   equ   $3C           ;lowest page
;Pages_Blk:     equ   $04           ;banks per block
C32:
eeprom_start:  equ   $0000         ;eeprom start (no EEprom)
eeprom_end:    equ   $0000         ;eeprom end  (no EEprom)
sector_size:   equ   $0200         ;Flash sector size
Flash_blks:    equ   $01           ;# of flash blks
lowest_page:   equ   $3E           ;lowest page
Pages_Blk:     equ   $02           ;pages per block
Dx32:
;eeprom_start:  equ   $0000         ;eeprom start (no EEprom)
;eeprom_end:    equ   $0000         ;eeprom end  (no EEprom)
;sector_size:   equ   $0200         ;Flash sector size
;Flash_blks:    equ   $01           ;# of flash blks
;lowest_page:   equ   $3E           ;lowest page
;Pages_Blk:     equ   $02           ;banks per block
;check above...

;*********************************************************************
;* One application specific Osc definition below should be uncommented
;*********************************************************************
;OscFreq:     equ   4000           ;Enter Osc speed
;initSYNR:    equ   $05            ; mult by synr + 1 = 6 (24MHz)
;initREFDV:   equ   $00            ;
;---------------------------------------------------------------------
OscFreq:     equ   8000           ;Enter Osc speed
initSYNR:    equ   $02            ; mult by synr + 1 = 3 (24MHz)
initREFDV:   equ   $00            ;
;---------------------------------------------------------------------
;OscFreq:     equ   16000          ;Enter Osc speed
;initSYNR:    equ   $02            ; mult by synr + 1 = 3/2 (24MHz)
;initREFDV:   equ   $01            ; divide by Refdiv + 1
;---------------------------------------------------------------------

;*********************************************************************
;* general equates for bootloader/monitor program valid for all
;* derivatives
;*********************************************************************
BootStart:   equ    $F800         ;start of protected boot block
RamStart:    equ    $2000         ;first RAM location
RamLast:     equ    $3fff         ;last RAM location (all devices)
Window:      equ    $8000         ;PPAGE Window start
RomStart:    equ    $4000         ;start of flash
VecTabSize:  equ    $80           ;size of vector table
VectorTable: equ    $10000-VecTabSize ;start of vector table
PVecTable:   equ    BootStart-VecTabSize ;start of pseudo vector table
FProtStart:  equ    $FF00         ;start of FLASH protection/security
;FProtBlksz:  equ    $C7           ;protect code for boot block ($C7 2K)
FProtBlksz:  equ    $FF           ;protect code for boot block (none)
FSecure:     equ    $BE           ;Disable Security and backdoor access

BusFreq:     equ    ((OscFreq/(initREFDV+1))*(initSYNR+1))
baud115200:  equ    (BusFreq/16)*10/1152  ;sets baud rate to 115,200
longBreak:   equ    1500          ;delay time for >30-bit break
; make TxD low at least 300us (30 bits @ 115200 baud)
; 5~ * 42ns/~ * 1500 = 315us (not exact, just >30 bit times)
asciiCR:     equ    $0D           ;ascii carriage return

flagReg:     equ    SCI0CR1   ;SCI control1 register associated with SCI0
RunFlag:     equ    WAKE      ;SCI Wake bit used as run/mon flag
Traceflag:   equ    ILT       ;SCI Idle bit used as trace flag
; 1=SWI caused by return from Trace1; 0=SWI from breakpoint or DBG

initSCI0CR2: equ    $0C     ;SCI0 Control Register 2
traceOne:    equ    $80     ;BKPCT0 pattern used for trace1 cmnd
;
;CPU HCS12 CCR immediately after rst is:
initUCcr:    equ    %11011000     ;initial value for user's CCR
;                    SX-HI---     ;I interrupts masked

ErrNone:     equ    $E0           ;code for no errors
ErrCmnd:     equ    $E1           ;command not recognized
ErrRun:      equ    $E2           ;command not allowed in run mode
ErrSP:       equ    $E3           ;SP was out of range
ErrWriteSP:  equ    $E4           ;attempted to write bad SP value
ErrByteNVM:  equ    $E5           ;write_byte attempt NVM
ErrFlash:    equ    $E6           ;FACCERR or FPVIOL error
ErrFlErase:  equ    $E7           ;improper flash erase range
ErrGoVec:    equ    $E8           ;Go or Trace1 with bad vectors

StatHalt:    equ    $02           ;stopped by Halt command
StatTrace:   equ    $04           ;returned from a Trace1 command
StatBreak:   equ    $06           ;Breakpoint or DBG (SWI) request
StatCold:    equ    $08           ;just did a cold reset
StatWarm:    equ    $0C           ;warm start because int with bad SP

;*********************************************************************
;* User CPU registers stack frame...
;*   +0  UCcr   <- Monitor's SP
;*   +1  UDreg   (B:A)
;*   +3  UXreg
;*   +5  UYreg
;*   +7  UPc
;*   +9  ---     <- User's SP
; Offsets from actual SP to user CPU regs while in monitor
;*********************************************************************

UCcr:        equ    0             ;user's CCR register
UDreg:       equ    1             ;user's D register (B:A)
UXreg:       equ    3             ;user's X register
UYreg:       equ    5             ;user's Y register
UPc:         equ    7             ;user's PC
SPOffset:    equ    9             ;offset of stack pointer while in monitor

MaxMonStack:  equ   35             ;maximum number of bytes used by Monitor
LowSPLimit:   equ   RamStart+MaxMonStack-SPOffset
HighSPLimit:  equ   RamLast-SPOffset+1

; named locations on stack if SWI with bad SP value
stackCheck:  equ    RamLast-2    ;SWI ISR leaves warmSig addr here
;*********************************************************************
;* Start of code and/or constant data
;*********************************************************************
             org    BootStart      ;beginning of protected flash
;*********************************************************************
;* Main startup program - real reset vector points to here
;*********************************************************************
;*********************************************************************
;  THIS LOCKS RAM / REGS / AND EEPROM BY USING ONLY WRITE USER CODE 
;  WILL NOT EFFECT THESE SETTINGS...This is for derivative compatibility
;  Please refer to EB388
;*********************************************************************
coldStart:  
Startup:
             movb  #$00,initrg    ;set registers at $0000 
             movb  #$39,initrm    ;set ram to end at $3fff 
             brclr  MEMSIZ0,eep_sw1+eep_sw0,coldStart1  ;Check if device has EEprom
             movb  #$09,initee    ;set eeprom to end at $0fff 
coldStart1:  ldaa  #StatCold      ;A = cold start flag test below

;**********************************************************************
; Decide whether to go to user reset or bootloader/monitor
;**********************************************************************
;  (a) default to monitor if high byte user pseudo-vector is erased ($FF)
;**********************************************************************
             ldab   vecReset-($10000-BootStart) ;check for user reset Vector
             comb                 ;if erased COMB result will be 0
             beq    Monitor       ;default to monitor if no vector

;**********************************************************************
; Test the state of some further pins to force entering the monitor
;   Depending on the hardware configuration enable/disable/modify the
;   sections below
;**********************************************************************
             bset   PUCR,PUPAE    ;enable pullup on force monitor sw
             bset   PPSS,PPSS0    ;enable pullup on RxD0 pin
             bset   PERS,PERS0    ;on RxD0 pin
             clrb
             dbne   b,*           ;delay to allow sw pin to pull up
;**********************************************************************
;  (b) force monitor if port A bit 6 = 01
;  Note: this port is configured after reset as input with pull-up
;**********************************************************************
             ldab   PORTA         ;get port value
             bitb   #PTA6         ;test the sw bit (A6)
             beq    Monitor

;**********************************************************************
;  (c) force monitor if RxD low (from host)  PORTS bit 0
;      This is true if the host holds RxD on break level
;  Note: this port is configured after reset as input with pull-up
;**********************************************************************
             brclr  PTS,PTS0,Monitor ;to monitor if RxD low
             bclr   PPSS,PPSS0    ;restore reset state on RxD0 pin
             bclr   PERS,PERS0    ;restore reset state on RxD1 pin

;**********************************************************************
;  finally jump to the user application (by pseudo vector)
;**********************************************************************
             jmp    [vecReset-($10000-BootStart),pcr] ;go where
                                        ;user reset vector points
;*********************************************************************
;  Fomal start of Monitor code
;*********************************************************************
Monitor:
; Initialize clock generator and PLL
;
             bclr    PUCR,PUPAE         ;restore reset state
             bclr    CLKSEL,PLLSEL      ;disengage PLL to system
             bset    PLLON,PLLSEL       ;turn on PLL
             movb    #initSYNR,SYNR     ;set PLL multiplier 
             movb    #initREFDV,REFDV   ;set PLL divider
             nop
             nop
             brclr   crgflg,lock,*+0    ;while (!(crg.crgflg.bit.lock==1))
             bset    CLKSEL,PLLSEL      ;engage PLL to system
;
; set flash/EEPROM clock to 200 kHz
;
             movb  #((OscFreq/200)-1),FCLKDIV ;Flash CLK = 200 kHz
             movb  #((OscFreq/200)-1),$110 ;Eeprom CLK = 200 kHz
;
; Set stack pointer to last (highest) RAM location
;
stackInit:   lds    #RamLast+1    ;point one past RAM
;
; set baud rate to 115.2 kbaud and turn on Rx and Tx
;
            movb  #baud115200,SCI0BDL  ;..BDH=0 so baud = 115.2 K
            movb  #initSCI0CR2,SCI0CR2 ;Rx and Tx on
                                  ;(SP will dec before store)
;
; if warm start, skip break and banner
; A is a flag to indicate cold vs warm start. avoid using A above
;
ChkCold:     cmpa  #StatCold     ;StatCold=cold start
             beq    coldBrk       ;if cold send break and banner
;
; Send a warm start prompt and wait for new commands
;
             ldaa  #ErrSP        ;error code for bad SP
             jsr    PutChar       ;send error code (1st prompt char)
             ldaa  #StatWarm     ;status code for warm start
             bra    endPrompt     ;finish warm start prompt
;
; Cold start so Generate long break to host
;
coldBrk:     brclr  SCI0SR1,TDRE,* ;wait for Tx (preamble) empty
             bset   SCI0CR2,SBK   ;start sending break after preamble
             ldx   #longBreak     ;at least 30 bit times for Windows
brkLoop:     cpx   #0             ;[2]done?
             dbne   x,brkLoop     ;[3]
             bclr   SCI0CR2,SBK   ;stop sending breaks

waitforCR:   jsr    GetChar       ;should be asciiCR or $00 with FE=1
             cmpa    #asciiCR      ;.eq. if 115.2K baud OK
             bne    waitforCR

; Setup initial user CPU register values (user register stack frame)
;   A holds the initial state value
initUregs:   ldx   BootStart-2    ;load user reset vector
             pshx                 ;$00 to user UPc      $3FFe
             ldx   #$0000
             pshx                 ;$00 to user UYreg    $3ffc
             pshx                 ;$00 to user UXreg    $3ffa
             pshx                 ;$00 to user UDreg    $3ff8
             ldab  #initUCcr      ;initial value for user CCR
             pshb                 ;to UCcr              $3ff7

;*********************************************************************
;* end of reset initialization, begin body of program
;*********************************************************************
;
; Send a cold start prompt and wait for new commands
;
             ldaa  #ErrNone       ;code for no errors ($E0)
             jsr    PutChar       ;send error code (1st prompt char)
             ldaa  #StatCold      ;status code for warm start
             bra    endPrompt     ;finish warm start prompt
;
; normal entry point after a good command
; Prompt is an alt entry point if an error occurred during a command
; endPrompt is an alternate entry for Trace1, Break (SWI), Halt,
; or warm/cold resets so an alternate status value can be sent
; with the prompt
;

commandOK:   ldaa  #ErrNone       ;code for no errors ($E0)
Prompt:      jsr    PutChar       ;send error code
             ldaa   flagReg       ;0 means monitor active mode
             anda  #RunFlag       ;mask for run/monitor flag (SCI WAKE)
             lsra                 ;shift flag to LSB
             lsra                 ;
             lsra                 ;$00=monitor active, $01=run
endPrompt:   jsr    PutChar       ;send status code
             ldaa  #'>'
             jsr    PutChar       ;send 3rd character of prompt seq

             brclr  flagReg,RunFlag,prompt1  ;no exit if run flag clr
             jmp    GoCmd         ;run mode so return to user program

prompt1:     jsr    GetChar       ;get command code character
             ldx   #commandTbl    ;point at first command entry
CmdLoop:     cmpa    ,x           ;does command match table entry?
             beq    doIt          ;branch if command found
             leax   3,x
             cpx   #tableEnd      ;see if past end of table
             bne    CmdLoop       ;if not, try next entry
             ldaa  #ErrCmnd       ;code for unrecognized command
             bra    Prompt        ;back to prompt; command error
             
doIt:        ldx    1,x           ;get pointer to command routine
             jmp     ,x           ;go process command
;
; all commands except GO, Trace_1, and Reset to user code - jump to
; Prompt after done. Trace_1 returns indirectly via a SWI.
;

;*********************************************************************
;* Command table for bootloader/monitor commands
;*  each entry consists of an 8-bit command code + the address of the
;*  routine to be executed for that command.
;*********************************************************************
commandTbl:  fcb   $A1
             fdb  RdByteCmd     ;read byte
             fcb   $A2
             fdb  WtByteCmd     ;write byte
             fcb   $A3
             fdb  RdWordCmd     ;read word of data 
             fcb   $A4
             fdb  WtWordCmd     ;write word of data 
             fcb   $A5
             fdb  RdNextCmd     ;read next word
             fcb   $A6
             fdb  WtNextCmd     ;write next word
             fcb   $A7
             fdb  RdNwordCmd    ;read n words of data
             fcb   $A8
             fdb  WtNwordCmd    ;write n words of data
             fcb   $A9
             fdb  RdRegsCmd     ;read CPU registers
             fcb   $AA
             fdb  WriteSpCmd    ;write SP
             fcb   $AB
             fdb  WritePcCmd    ;write PC
             fcb   $AC
             fdb  WriteIYCmd    ;write IY
             fcb   $AD
             fdb  WriteIXCmd    ;write IX
             fcb   $AE
             fdb  WriteDCmd     ;write D
             fcb   $AF
             fdb  WriteCcrCmd   ;write CCR
             fcb   $B1
             fdb  GoCmd         ;go
             fcb   $B2
             fdb  Trace1Cmd     ;trace 1
             fcb   $B3
             fdb  HaltCmd       ;halt
             fcb   $B4
             fdb  ResetCmd      ;reset - to user vector or monitor
;            Command not implemented
             fcb   $B6          ;code - erase flash command
             fdb  EraseAllCmd   ;erase all flash and eeprom command routine
             fcb   $B7          ;return device ID
             fdb  DeviceCmd
             fcb   $B8          ;erase current flash bank selected in PPAGE
             fdb  ErsPage
tableEnd:    equ    *           ;end of command table marker

;*********************************************************************
;* Device ID Command -  Ouputs hex word from device ID register
;*********************************************************************
DeviceCmd:
             ldaa   #$DC         ;get part HCS12 descripter
             jsr    PutChar      ;out to term
             ldaa   PARTIDH     ;get part ID high byte
             jsr    PutChar      ;out to term
             ldaa   PARTIDL     ;get part ID low byte
             jsr    PutChar      ;out to term
             ldaa    #ErrNone    ;error code for no errors
             jmp    Prompt       ;ready for next command


;*********************************************************************
;* Halt Command - halts user application and enters Monitor
;*   This command is normally sent by the debugger while the user
;*   application is running. It changes the state variable in order
;*   to stay in the monitor
;*********************************************************************
HaltCmd:     bclr   flagReg,RunFlag ;run/mon flag = 0; monitor active
             ldaa    #ErrNone      ;error code for no errors
             jsr    PutChar        ;send error code
             ldaa    #StatHalt     ;status code for Halt command
             jmp    endPrompt      ;send status and >
;*********************************************************************
;* Halt user code by Rx interrupt of SCI
;*
;*********************************************************************
uSci0:       bset   flagReg,RunFlag  ;set run/mon flag (run mode)
             jmp    prompt1       ;go get and decode command
;* unlike most ISRs, this one does not end in an RTI. If/when we
;* return to running the user program, we will re-enable Rx interrupts

;*********************************************************************
;* Reset Command - forces a reset - if user pseudo-vector is not blank
;*  (or some other conditions are met - see coldstart:) processing will
;*  start at the user-specified reset pseudo-vector location and the
;*  user has full control of all write-once registers. Otherwise reset
;*  causes the bootloader/monitor program to cold start.
;*********************************************************************
ResetCmd:     movb  #StatCold,stackCheck;force cold startup (see Startup)
              ldaa  #RSBCK|!CR2|!CR1|CR0 ;Cop disabled in BDM
              staa   COPCTL        ; turn on cop monitor
              cmpa   COPCTL        ; load to see if user touched it
              beq    CopLock       ; wait for COP reset
              jmp    coldstart     ; can't use COP just start over
CopLock:      bra	 *

;*********************************************************************
;* SWI service routine - trace1 or breakpoint from user code
;*  SWI saves user CPU registers on stack and returns to monitor
;*  control at a new command prompt.
;* User CPU registers stack frame...
;*
;*   +0  UCcr	<- Monitor's SP
;*   +1  UDreg   (B:A)
;*   +3  UXreg
;*   +5  UYreg
;*   +7  UPc
;*   +9  ---     <- User's SP
;*********************************************************************
Breakpoint:  clr    $28            ;Bkpct0 cleared to disabled
             bclr   flagReg,RunFlag  ;run/mon flag = monitor active
             ldaa   #StatTrace     ;set status to Trace
                                   ;and enter monitor

;*********************************************************************
;* This is the entrypoint to the monitor from the user application
;*   A contains the status value that reflects run status
;*
;* If SP isn't within valid RAM, it can't support the monitor so the
;* monitor is forced to initialize the SP and user registers.
;*********************************************************************
ReenterMon:  cps   #LowSPLimit+1  ;check against lower limit
             blo    badSP         ; note: +1 => A is not pushed yet
             cps   #HighSPLimit+1 ;check against upper limit
             bhi    badSP
             ldaa    #ErrNone      ;error code for no errors
             jsr    PutChar       ;send error code
             ldaa    #StatBreak    ;status code for Break (SWI)
             brclr  flagReg,TraceFlag,SWIdone  ;0 indicates not Trace
             bclr   flagReg,TraceFlag  ;acknowledge trace flag
             ldaa    #StatTrace    ;status code for Trace1 return
SWIdone:     jmp    endPrompt     ;send status and >
badSP:       ldaa   #ErrSP        ;set error code to bad stack pointer
             jmp    stackInit

;*********************************************************************
;* Erase Command - Use repeated page erase commands to erase all flash
;*  except bootloader in protected block at the end of flash, and mass
;*  erase all EEPROM locations
;*
;* Eeprom erasure assumes no protection. (Mass command will fail)
;*********************************************************************
EraseAllCmd: jsr    abClr         ;abort commands and clear errors

             brclr  MEMSIZ0,eep_sw1+eep_sw0,ErsBlk0  ;Check if device has EEprom
             ldy    #eeprom_start ; get device eeprom start
             std    0,y           ; write to eeprom (latch address)
                                  ; data is don't care (but needed)

             movb  #$41,$116      ;mass erase command
             movb  #CBEIF,$115   ;register the command
             nop                  ; wait a few cycles for
             nop                  ; command to sync.
             nop
ChkDoneE:    ldaa   $115         ;wait for CBEIF=CCIF=1 (cmnd done)
             bpl    ChkDoneE      ;loop if command buffer full (busy)
             asla                 ;moves CCIF to MSB (set/clear N bit)
             bpl    ChkDoneE      ;loop if CCIF=0 (not done)
;
; erase flash pages from RomStart to start of protected bootloader
; no need to check for errors because we cleared them before EE erase
;

ErsBlk0:                          ; sector erase all full blocks
             ldab   #Pages_Blk    ; Get number of banks/blocks
             decb                 ; erase all but last
             stab   1,-sp         ; save counter
             ldaa   #$3f          ; highest bank
             sba                  ; Compute lowest page-1
             staa   PPAGE         ; PPAGE for first 16K page of block 0
                                  ; (passed in the A accumulator).
             clr    FCNFG         ; set block select bits to 0.
ErsBlk0Lp:   ldx    #sector_size  ; select sector size
             ldd    #$4000        ; Window size
             idiv                 ; compute total number of sectors
             tfr    x,d           ; get number of sectors in B
             ldx   #Window        ; point to the start of the PPAGE window.
             bsr    ErsSectors    ; go erase the PPAGE window a sector at a time.
             inc    PPAGE         ; go to the next PPAGE.
             dec    0,sp          ; done with all full PPAGE blocks?
             bne    ErsBlk0Lp     ;   no? then erase more blocks.

             ldx    #sector_size  ; select sector size
             ldd    #((BootStart-$c000)) ; get size - protected amount
             idiv                 ; compute total number of sectors
                                  ; minus the bootblock.
             tfr    x,d           ; get number of sectors in B
             ldx   #Window        ; point to the start of the PPAGE window.
             bsr    ErsSectors    ; go erase the PPAGE window a sector at a time.
             pulb                 ; remove the page count from the stack.

; erase all sectors outside the bootblock.
;
;********************************************************************
;bulk erase all the rest
;********************************************************************

             ldab  #Flash_blks    ; select lowest page of the highest bank
             decb                 ;
             beq    EraseDone     ; if single block device quit
             ldab  #lowest_page   ; select lowest bank
BlockLoop:   stab   ppage         ; must match array selection
             lsrb                 ; calculate the value of the block select bits based
             lsrb                 ; on bits 3:2 of the PPAGE register value. (<256k)
             ldy   #sector_size   ; get high byte of size
             cpy   #$0200         ; if larger than $200 shift again
             beq    nBlockLoop    ; otherwise skip ahead
             lsrb                 ; on bits 4:3 of the PPAGE register value. (512k)
nBlockLoop:  comb
             andb  #$03           ; mask off all but the lower 2 bits.
             stab   FCNFG         ; select the block to erase.
             bsr    BulkErase     ; erase it
             ldab   ppage         ;get ppage back
             addb  #Pages_Blk     ;
             cmpb  #($3F-Pages_Blk) ; see if last block
             bmi    BlockLoop

EraseDone:   movb  #$3D,ppage     ;select bank in array0

OkCommand:   jmp    commandOK     ;back to no error and prompt

BulkErase:   pshx                 ;save address
             ldx    #Window       ;must point into bank
             staa   ,x            ;latch address to erase
             movb  #$41,FCMD      ; Select mass erase
             movb  #CBEIF,FSTAT   ;register the command
             nop                  ;wait a few cycles for
             nop                  ;command to sync.
             nop
ChkDoneF:    ldaa   FSTAT         ;wait for CBEIF=CCIF=1 (cmnd done)
             bpl    ChkDoneF      ;loop if command buffer full (busy)
             asla                 ;moves CCIF to MSB (set/clear N bit)
             bpl    ChkDoneF      ;loop if CCIF=0 (not done)
             pulx                 ;get address back
             rts
;Erase 'b' (accumulator) sectors beginning at address 'x' (index register)
;
ErsSectors:  exg    b,y           ;put the sector count in y.
ErsSectLp:   std    ,x
             movb  #$40,FCMD      ;perform a sector erase.
             jsr    DoOnStack     ;finish command from stack-based sub
             tsta                 ;check for 0=OK
             bne    ErsSectErr    ;back to prompt-flash erase error
             leax   sector_size,x ;point to the next sector.
             dbne   y,ErsSectLp   ;continue to erase remaining sectors.
             rts

ErsSectErr:  puld                 ; clear stack
             bra    ErsPageErr

ErsPage:     jsr    abClr         ; abort commands and clear errors
	       ldab   PPAGE         ; get current ppage

             lsrb                 ; calculate the value of the block select bits based
             lsrb                 ; on bits 3:2 of the PPAGE register value. (<256k)
             ldy   #sector_size   ; get high byte of size
             cpy   #$0200         ; if larger than $200 shift again
             beq    ErsPage1      ; otherwise skip ahead
             lsrb                 ; on bits 4:3 of the PPAGE register value. (512k)
ErsPage1:    comb
             andb  #$03           ; mask off all but the lower 2 bits.
             stab   FCNFG         ; select the block to erase.
             ldab   PPAGE         ; get current ppage
             cmpb  #$3F		      ; is it the page with the monitor
             bne   ErsFullPage    ; no then erase all of page
             ldx   #sector_size   ; select sector size
             ldd   #((BootStart-$c000)) ; get size - protected amount
             idiv                 ; compute total number of sectors
                                  ; minus the bootblock.
             tfr    x,d           ; get number of sectors in B
             ldx   #Window        ; point to the start of the PPAGE window.
             bsr    ErsSectors    ; go erase the PPAGE window a sector at a time.
             bra    EraPageStat   ; back to no error and prompt

ErsFullPage: ldx   #sector_size   ; select sector size
             ldd   #$4000         ; Window size
             idiv                 ; compute total number of sectors
             tfr    x,d           ; get number of sectors in B
             ldx   #Window        ; point to the start of the PPAGE window.
             bsr    ErsSectors    ; go erase the PPAGE window a sector at a time.
             bra    EraPageStat     ;back to no error and prompt

EraPageStat: ldaa   FSTAT
             anda  #$30           ;mask all but PVIOL or ACCERR
             bne    ErsPageErr    ;back to prompt-flash error
             ldaa   #ErrNone      ;code for no errors ($E0)
             jmp    Prompt        ;ready for next command

ErsPageErr: ldaa   #ErrFlash     ;code for Flash error ($E6)
             jmp    Prompt        ;ready for next command

;*********************************************************************
;* Read Byte Command - read specified address and return the data
;*  8-bit command code from host to SCI0 RxD
;*  16-bit address (high byte first) from host to SCI0 RxD
;*  8-bit data sent back to host through SCI0 TxD
;*********************************************************************
RdByteCmd:   jsr    getX          ;get address to read from
             ldaa   ,x            ;read the requested location
             jsr    PutChar       ;send it out SCI0
             jmp    commandOK     ;ready for next command

;*********************************************************************
;* Read Word Command - read specified block of data
;*  8-bit command code from host to SCI0 RxD
;*  16-bit address (high byte first) from host to SCI0 RxD
;*  16-bit number sent back to host through SCI0 TxD
;* Special case of block read.
;*********************************************************************
RdWordCmd:   jsr    getX          ;get address to read from
sendExit:    ldd    ,x            ;read the requested location
             jsr    PutChar       ;send it out SCI0
             tba
             jsr    PutChar       ;send it out SCI0
             jmp    commandOK     ;ready for next command

;*********************************************************************
;* Read Block Command - read specified block of data
;*  8-bit command code from host to SCI0 RxD
;*  16-bit address (high byte first) from host to SCI0 RxD
;*  8-bit number of words-1 to sent back to host through SCI0 TxD
;*********************************************************************
RdNwordCmd:  jsr    getX          ;get address to read from
             jsr    GetChar       ;get number of bytes to read
             tab
             incb                 ;change 0-255 to 1-256(00)
             pshb			      ;save count
ReadWDNext:  ldd    2,x+            ;read the requested location
;             inx
;             inx
             jsr    PutChar       ;send it out SCI0
             tba
             jsr    PutChar       ;send it out SCI0
			 dec    ,sp           ;dec counter on stack
             bne    ReadWDNext
             pulb                 ;restore count
             ldaa    #ErrNone     ;code for no errors ($E0)
             jmp    Prompt        ;ready for next command

;*********************************************************************
;* Read Next Command - IX=IX+2; read m(IX,IX=1) and return the data
;*  8-bit command code from host to SCI0 RxD
;*  16-bit data sent back to host through SCI0 TxD
;*  uses current value of IX from user CPU regs stack frame
;*********************************************************************
RdNextCmd:   brclr  flagReg,RunFlag,notRun  ;do command if not run
             clra                 ;data = $00 (can't read real data)
             jsr    PutChar       ;send $00 instead of read_next data
             jsr    PutChar       ;send $00 instead of read_next data
             ldaa   #ErrRun       ;code for run mode error
xCmnd:       jmp    Prompt        ;back to prompt; run error
notRun:      bsr    preInc        ;get, pre-inc, & update user IX
             bra    sendExit      ;get data, send it, & back to prompt

;*********************************************************************
;* Write Byte Command - write specified address with specified data
;*  8-bit command code from host to SCI0 RxD
;*  16-bit address (high byte first) from host to SCI0 RxD
;*  8-bit data from host to SCI0 RxD
;*********************************************************************
WtByteCmd:   jsr    getX          ;get address to write to
WriteNext2:  jsr    GetChar       ;get data to write
             jsr    CheckModule
             beq    isRAMbyte
             bra    WriteByteNVM  ;deny access (byte NVM access)

isRAMbyte:   staa   0,x           ;write to RAM or register
             clra                 ;force Z=1 to indicate OK

WriteExit:   ldaa   #ErrNone      ;code for no errors ($E0)
             jmp    Prompt        ;ready for next command

WriteByteNVM: ldaa   #ErrByteNVM     ;code for byte NVM error ($E5)
             jmp    Prompt        ;ready for next command


;*********************************************************************
;* Write Word Command - write word of data
;*  8-bit command code from host to SCI0 RxD
;*  16-bit address (high byte first) from host to SCI0 RxD
;*  16-bit value to write
;*********************************************************************
WtWordCmd:   jsr    getX          ;get address to write to
			 ldab	#01           ;one word
             bra    WriteWord     ;get & write data, & back to prompt
             
;*********************************************************************
;* Write Block Command - write specified block of data
;*  8-bit command code from host to SCI0 RxD
;*  16-bit address (high byte first) from host to SCI0 RxD
;*  8-bit number of words-1 to write from host to SCI0 TxD
;*  n 16-bit values to write
;*
;*********************************************************************
WtNwordCmd:  jsr    getX          ;get address to write to
             jsr    GetChar       ;get number of bytes to read
             tab
             incb                 ;change 0-255 to 1-256(00)
WriteWord:   psha                 ;decrements stack by one
             clr     ,sp          ;clear error flag on stack
             pshb                 ;preserve byte counter
             jsr    GetChar       ;get high byte
             tab                  ;save in B
             jsr    GetChar       ;get low byte
             exg    a,b           ;flip high and low byte
             jsr    WriteD2IX     ;write or program data to address
             pulb                 ;restore byte counter
             beq    WriteOK       ;error detected
             ldaa    #1           ;a non-0 value means some error
             staa    ,sp          ;update error flag on stack
WriteOK:     inx                  ;increment
             inx				  ; 2x
             decb				  ; count -1
             bne    WriteWord
             pula                 ;recover error flag
             tsta                 ;check for 0=OK
             bne    WriteError    ;back to prompt-flash error
             
             ldaa   #ErrNone      ;code for no errors ($E0)
             jmp    Prompt        ;ready for next command

WriteError:  ldaa   #ErrFlash     ;code for Flash error ($E6)
             jmp    Prompt        ;ready for next command

;*********************************************************************
;* Write Next Command - IX=IX+1; write specified data to m(IX)
;*  8-bit command code from host to SCI0 RxD
;*  16-bit data from host to SCI0 RxD
;*
;*  uses current value of IX from user CPU regs stack frame
;*********************************************************************
WtNextCmd:   brclr  flagReg,RunFlag,notRunW  ;do command if not run
             jsr    getX          ;clear data
             ldaa   #ErrRun       ;code for run mode error
xCmndW:      jmp    Prompt        ;back to prompt; run error

notRunW:     bsr    preInc        ;get, pre-inc, & update user IX
			 ldab	#01           ;one word
             bra    WriteWord     ;get & write data, & back to prompt

;*********************************************************************
;* utility to get IX from stack frame and pre increment it by 2
;* assumes interrupts are blocked while in monitor
;*********************************************************************
preInc:      leas 2,sp
             ldx    UXreg,sp      ;get user X
             inx                  ;pre-increment
             inx                  ;pre-increment
             stx    UXreg,sp      ;put adjusted user X back on stack
             leas -2,sp
             rts                  ;pre-incremented IX still in IX

;*********************************************************************
;* Read Registers Command - read user's CPU register values
;*
;*  16-bit SP value (high byte first) sent to host through SCI0 TxD
;*  16-bit PC value (high byte first) sent to host through SCI0 TxD
;*  16-bit IY value (high byte first) sent to host through SCI0 TxD
;*  16-bit IX value (high byte first) sent to host through SCI0 TxD
;*  16-bit D  value (high byte first) sent to host through SCI0 TxD
;*   8-bit CCR value sent to host through SCI0 TxD
;*
;* User CPU registers stack frame...
;*
;*   +0  UCcr   <- Monitor's SP
;*   +1  UDreg   (B:A)
;*   +3  UXreg
;*   +5  UYreg
;*   +7  UPc
;*   +9  ---     <- User's SP
;*********************************************************************
RdRegsCmd:   tsx                  ;IX = Monitor SP +2
             leax   SPOffset,x    ;correct SP value
             jsr    put16         ;send user SP out SCI0
             ldx    UPc,sp        ;user PC to IX
             jsr    put16         ;send user PC out SCI0
             ldx    UYreg,sp      ;user IY to IX
             jsr    put16         ;send user IY out SCI0
             ldx    UXreg,sp      ;user IX to IX
             jsr    put16         ;send user IX out SCI0
             ldx    UDreg,sp      ;user D to IX
             exg    d,x
             exg    a,b           ;flip as D is stacked B:A
             exg    d,x
             jsr    put16         ;send user D out SCI0
             ldaa   UCcr,sp       ;user CCR to A
             jsr    PutChar       ;send user CCR out SCI0
             jmp    CommandOK     ;back to prompt

;*********************************************************************
;* Write CCR Command - write user's CCR register value
;*  8-bit command code from host to SCI0 RxD
;*  8-bit data for CCR from host to SCI0 RxD
;*********************************************************************
WriteCcrCmd: jsr    GetChar       ;read new CCR value
             staa   UCcr,sp       ;replace user CCR value
             jmp    commandOK     ;back to no error and prompt

;*********************************************************************
;* Write D Command - write user's D register value
;*  8-bit command code from host to SCI0 RxD
;*  16-bit data (high byte first) for D from host to SCI0 RxD
;*********************************************************************
WriteDCmd:   jsr    getX          ;read new D value
             exg    d,x
             exg    a,b           ;flip as D is stacked B:A
             exg    d,x
             stx    UDreg,sp      ;replace user D value
             jmp    commandOK     ;back to no error and prompt

;*********************************************************************
;* Write IX Command - write user's IX register value
;*  8-bit command code from host to SCI0 RxD
;*  16-bit data (high byte first) for IX from host to SCI0 RxD
;*********************************************************************
WriteIXCmd:  bsr    getX          ;read new IX value
             stx    UXreg,sp      ;replace user IX value
             jmp    commandOK     ;back to no error and prompt

;*********************************************************************
;* Write IY Command - write user's IY register value
;*  8-bit command code from host to SCI0 RxD
;*  16-bit data (high byte first) for IY from host to SCI0 RxD
;*********************************************************************
WriteIYCmd:  bsr    getX          ;read new IY value
             stx    UYreg,sp      ;replace user IY value
             jmp    commandOK     ;back to no error and prompt

;*********************************************************************
;* Write PC Command - write user's PC register value
;*  8-bit command code from host to SCI0 RxD
;*  16-bit data (high byte first) for PC from host to SCI0 RxD
;*********************************************************************
WritePcCmd:  bsr    getX          ;read new PC thru SCI0 to IX
             stx    UPc,sp       ;replace user PC value
             jmp    commandOK     ;back to no error and prompt

;*********************************************************************
;* Write SP Command - write user's SP register value
;*  8-bit command code from host to SCI0 RxD
;*  16-bit data (high byte first) for SP from host to SCI0 RxD
;*
;*  Since other user CPU register values are stored on the stack, the
;*  host will need to re-write the other user registers after SP is
;*  changed. This routine just changes SP itself.
;*
;*  SP value is user's SP & it is adjusted (-10) to accommodate the
;*  user CPU register stack frame.
;*
;*  If the host attempts to set the user SP value <RamStart or >RamLast
;*  then the change is ignored, because such values would not support
;*  proper execution of the monitor firmware.
;*********************************************************************

;??? why does P&E need opp. (leax  -SPOffset,x) ? 
WriteSpCmd:  bsr    getX         ;new SP value now in IX
             leax  SPOffset,x    ;correct SP value
             cpx   #LowSPLimit   ;check against lower limit
             blo    spBad
             cpx   #HighSPLimit  ;check against upper limit
             bhi    spBad
             txs                 ;IX -> SP
             jmp    commandOK    ;back to no error and prompt
spBad:       ldaa    #ErrSP      ;error code for stack errors
             bsr    PutChar      ;send error code
             jmp    Prompt       ;send status and >

;*********************************************************************
;* Trace 1 Command - trace one user instruction starting at current PC
;*  8-bit command code from host to SCI0 RxD
;*
;*  if an interrupt was already pending, the user PC will point at the
;*  ISR after the trace and the opcode at the original address will
;*  not have been executed. (because the interrupt response is
;*  considered to be an instruction to the CPU)
;*********************************************************************
;
pagebits:   fcb      $3D
            fcb      $3E
            fcb      $3F
pagebitsaddr:

            fdb     pagebits      ;$0000-$3FFF : Use constant $3D
            fdb     pagebits+1    ;$4000-$7FFF : Use constant $3E (2nd last page)
            fdb     $0030         ;$8000-$BFFF : Use PPAGE
            fdb     pagebits+2    ;$C000-$FFFF : Use constant $3F (last page)
Trace1Cmd:
            bset   flagReg,TraceFlag  ;so at SWI we know it was Trace
            ldx     UPc,sp       ;PC of go address
            inx                  ;IX points at go opcode +1
            inx                  ;IX points at go opcode +2
 	      xgdx
            andb   #$FE
            std     $2B          ;debugger trigger address
            std     $2E          ;same for second address to have it initialized
            rola
            rolb
            rola
            rolb
            andb   #$03          ;what range 0-3fff,4000-7fff,8000-bfff,c000-ffff?
            clra
            lsld
            xgdx
            ldx     pagebitsaddr,X
            ldaa    0,x
            staa    $2A          ;set page byte of address
            staa    $2D          ;same for second address to have it initialized
            ldaa   #traceOne     ;enable, arm, CPU force
            staa    $28          ;arm DBG to trigger after 1 instr.
            rti                  ;restore regs and go to user code

;*********************************************************************
;* Go Command - go to user's program at current PC address
;*  8-bit command code from host to SCI0 RxD
;* - no promt is issued 
;*  typically, an SWI will cause control to pass back to the monitor
;*********************************************************************
GoCmd:       bset SCI0CR2,RIE     ;need to enable SCI0 Rx interrupts to
                                  ; enter monitor on any char received
             bclr  flagReg,TraceFlag ; run flag clr
             rti                  ;restore regs and exit
;*********************************************************************
;* Utility subroutines - usually placed after main code
;*********************************************************************

;*********************************************************************
;* Utility to send a 16-bit value out X through SCI0
;*********************************************************************
put16:       exg    d,x           ;move IX to A
             bsr    PutChar       ;send high byte
             tba                  ;move B to A
             bsr    PutChar       ;send low byte
             rts

;*********************************************************************
;* Utility to get a 16-bit value through SCI0 into X
;*********************************************************************
getX:        bsr    GetChar       ;get high byte
             tab                  ;save in B
             bsr    GetChar       ;get low byte
             exg    a,b           ;flip high and low byte
             exg    d,x           ;16-bit value now in IX
             rts
;*********************************************************************
;* GetChar - wait indefinitely for a character to be received
;*  through SCI0 (until RDRF becomes set) then read char into A
;*  and return. Reading character clears RDRF. No error checking.
;*
;* Calling convention:
;*            bsr    GetChar
;*
;* Returns: received character in A
;*********************************************************************
GetChar:     brset  SCI0SR1,RDRF,RxReady ;exit loop when RDRF=1
             bra    GetChar              ;loop till RDRF set
RxReady:     ldaa   SCI0DRL              ;read character into A
             rts                         ;return

;*********************************************************************
;* PutChar - sends the character in A out SCI0
;*
;* Calling convention:
;*            ldaa    data          ;character to be sent
;*            bsr    PutChar
;*
;* Returns: nothing (A unchanged)
;*********************************************************************
PutChar:     brclr   SCI0SR1,TDRE,PutChar ;wait for Tx ready
             staa    SCI0DRL       ;send character from A
             rts

;********************************************************************
;*  PutHexWd -
;*  convert word pointed to by IX from hexadecimal to a ASCII number
;*  outputted to SCI0
;*********************************************************************
;PutHexWd:    psha
;             ldaa   0,X           ;get data in a
;             bsr    PutHexB       ;output left half
;             ldaa   1,X           ;get data in a
;             bsr    PutHexB       ;output right half
;             pula
;             rts
;
;********************************************************************
;*  PutHexB -
;*  convert byte pointed to by IX from hexadecimal to a ASCII number
;*  outputted to SCI0
;*********************************************************************
;PutHexB:     psha                 ;save copy
;             lsra                 ;shift data to right
;             lsra
;             lsra
;             lsra
;             bsr    Hex2asc       ;output left half
;             pula                 ;retrieve copy
;             bsr    Hex2asc       ;output right half
;             rts
;
;********************************************************************
;*  Hex2asc -
;*  convert nibble in low half of A from hexadecimal to a ASCII
;*  number outputted to SCI0
;*********************************************************************
;Hex2asc:      anda  #$0F           ;mask top half
;             adda  #$30           ;convert to ascii
;             cmpa  #$39
;             ble    GoOut         ;jump if 0-9
;             adda  #$07           ;convert to hex A-F
;GoOut:       bsr    PutChar       ;output character
;             rts

;*********************************************************************
;* CheckModule - check in what memory type the address in IX points to
;*  The location may be RAM, FLASH, EEPROM, or a register
;*  if the vector table is addresses, IX is changed to point to the
;*  same vector in the pseudo vector table
;*  returns in B: 1 FLASH or EEPROM
;*                0 RAM or register (all the rest of the address space)
;*               -1 access denied (monitor or pseudo vector)
;*  all registers are preserved except B
;*********************************************************************
CheckModule: pshd                 ;preserve original data
             cpx    #RomStart
             blo    check4EE      ;skip if not flash
             cpx    #VectorTable
             bhs    isVector      ;is it in the real vector table
             cpx    #PVecTable
             blo    isToProgram   ;pseudo vector table or monitor area
             ldab   #$FF          ;access denied (N=1, Z=0)
             puld                 ;restore original data (D)
             rts

isVector:    leax   BootStart,x   ;access pseudo vector table
             bra    isToProgram

check4EE:    cpx   #eeprom_start
             blo    isRAM         ;treat as RAM or registers
             cpx   #eeprom_end
             bhi    isRAM         ;must be high-page registers
isToProgram: ldab   #1            ;set flgs - signal FLASH (N=0, Z=0)
             puld                 ;restore original data (D)
             rts

isRAM:       clrb                 ;signal RAM  (N=0, Z=1)
             puld                 ;restore original data (D)
             rts

;*********************************************************************
;* WriteD2IX - Write the data in D (word) to the address in IX
;*  The location may be RAM, FLASH, EEPROM, or a register
;*  if FLASH or EEPROM, the operation is completed before return
;*  IX and A preserved, returns Z=1 (.EQ.) if OK
;*
;*********************************************************************
WriteD2IX:   pshx                 ;preserve original address
             pshd                 ;preserve original data
             bsr    CheckModule
             bmi    ExitWrite     ;deny access (monitor or pseudo vector)
             beq    isRAMword
             cpd    0,x           ;FLASH or EEPROM needs programming
             beq    ExitWrite     ;exit (OK) if already the right data
             pshd                 ;temp save data to program
             tfr    x,b           ;low byte of target address -> B
             bitb   #1            ;is B0 = 1?
             bne    oddAdrErr     ;then it's odd addr -> exit
             ldd    0,x           ;$FF if it was erased
             cpd    #$FFFF        ;Z=1 if location was erased first
oddAdrErr:   puld                 ;recover data, don't change CCR
             bne    ExitWrite     ;exit w/ Z=0 to indicate error
             bra    DoProgram

isRAMword:   std    0,x           ;write to RAM or register
             clra                 ;force Z=1 to indicate OK
             bra    ExitWrite

; Programs D to IX in either FLASH or EEPROM
DoProgram:   bsr    abClr         ;abort commands and clear errors
             cpx    #RomStart     ;simple test only
             blo    itsEE         ; details already verified
             bsr    ProgFlash     ;program the requested location
             bra    ExitWrite     ;exit (Z indicates good or bad)
itsEE:       bsr    ProgEE        ;program the requested location
; exit WriteD2IX functions (Z indicates good or bad)
ExitWrite:   puld                 ;restore original data (D)
             pulx                 ;restore original address (IX)
             rts

;*********************************************************************
;* Progee - program a single word in the HCS9S12 EEPROM
;*  the location is assumed to be previously erased. This routine
;*  waits for the command to complete.
;*
;* On entry... IX - points at the EEPROM address to be programmed
;*  A - holds the data value to be programmed
;*
;* Calling convention:
;*           bsr    Prog1ee
;*
;* Returns: IX unchanged and A = ESTAT shifted left by 2 bits
;*  Z=1 if OK, Z=0 if protect violation or access error
;*********************************************************************
ProgEE:      std     ,x           ;latch address & data to program
             ldaa    #ProgWord    ;Select program word command
             staa    $116         ;issue word program command
             ldaa    #CBEIF
             staa    $115        ;[pwpp] register command
             nop                  ;[p]
             nop
             nop
ChkDoneEE:   ldaa    $115        ;[prpp] (min 4~ before 1st read)
             anda   #$C0          ; mask all but 2 msb
             lsla                 ;CCIF now in MSB
             bpl    ChkDoneEE     ;wait for queued commands to finish
             asla                 ;A=00 & Z=1 unless PVIOL or ACCERR
xProgEE:     rts

;
; utility sub to abort previous commands in flash and EEPROM
; and clear any pending errors
;
abClr:       psha
             ldaa    #PVIOL+ACCERR ;mask
             staa    $115         ;abort any command and clear errors
             staa    FSTAT         ;abort any command and clear errors
             pula
             rts

;*********************************************************************
;* Progflash - programs one byte of HCS9S12 FLASH
;*  This routine waits for the command to complete before returning.
;*  assumes location was blank. This routine can be run from FLASH
;*
;* On entry... IX - points at the FLASH byte to be programmed
;*             A holds the data for the location to be programmed
;*
;* Calling convention:
;*           bsr    Prog1flash
;*
;* Uses: DoOnStack which uses SpSub
;* Returns: IX unchanged and A = FSTAT bits PVIOL and ACCERR only
;*  Z=1 if OK, Z=0 if protect violation or access error
;*********************************************************************
ProgFlash:   pshd
             cpx   #$8000         ; if <$8000 then bank 3E
             blo    its3E         ;set ppage to 3E
             cpx   #$C000         ; if > $BFFF then bank 3F
             blo    ProgFlash1    ;set ppage 3F
             movb  #$3F,ppage     ;
             bra   ProgFlash1
its3E:       movb  #$3E,ppage     ;

ProgFlash1:  ldab   PPAGE
             lsrb                 ; calculate the value of the block select bits based
             lsrb                 ; on bits 3:2 of the PPAGE register value. (<256k)
             ldy   #sector_size   ; get high byte of size
             cpy   #$0200         ; if larger than $200 shift again
             beq    nBlockLoopb
             lsrb                 ; on bits 4:3 of the PPAGE register value. (512k)

nBlockLoopb: comb
             andb  #$03           ; mask off all but the lower 2 bits.
             stab   FCNFG         ; select the block to program.
             cmpb  #$00           ; if block zero use DoOnStack method
             puld
             beq    ProgFlashSP

ProgFlshRom: std    ,x            ;latch address & data to program
             ldaa  #ProgWord         ;Select program word command
             staa   FCMD          ;issue byte program command
             ldaa  #CBEIF
             bsr    SpSub         ;register command & wait to finish
             ldaa   FSTAT
             anda  #$30           ;mask all but PVIOL or ACCERR
             rts

ProgFlashSP: std    ,x            ;latch address and data
             ldaa  #ProgWord         ;Select program word command
             staa   FCMD          ;issue byte program command
;
; DoOnStack will register the command then wait for it to finish
;  in this unusual case where DoOnStack is the next thing in program
;  memory, we don't need to call it. The rts at the end of DoOnStack
;  will return to the code that called Prog1flash.
;
;*********************************************************************
;* DoOnStack - copy SpSub onto stack and call it (see also SpSub)
;*  De-allocates the stack space used by SpSub after returning from it.
;*  Allows final steps in a flash prog/erase command to execute out
;*  of RAM (on stack) while flash is out of the memory map
;*  This routine can be used for flash byte-program or erase commands
;*
;* Calling Convention:
;*           bsr    DoOnStack
;*
;* Uses 16 bytes on stack + 2 bytes if BSR/bsr used to call it
;* returns IX unchanged
;********************************************************************
DoOnStack:   pshx                 ;save IX
             ldx   #SpSubEnd      ;point at last byte to move to stack
SpmoveLoop:  ldaa    ,x           ;read from flash
             psha                 ;move onto stack
             dex                  ;next byte to move
             cpx   #SpSub-1       ;past end?
             bne    SpmoveLoop    ;loop till whole sub on stack
             tfr   sp,x           ;point to sub on stack
             ldaa  #CBEIF         ;preload mask to register command
             jsr    ,x            ;execute the sub on the stack
             leas   SpSubEnd-SpSub+1,sp  ;de-allocate space used by sub
             ldaa   FSTAT         ;get result of operation
             anda  #$30           ;and mask all but PVIOL or ACCERR
             pulx                 ;restore IX
             rts                  ;to flash where DoOnStack was called

;*********************************************************************
;* SpSub - register flash command and wait for Flash CCIF
;*  this subroutine is copied onto the stack before executing
;*  because you can't execute out of flash while a flash command is
;*  in progress (see DoOnStack to see how this is used)
;*
;* Uses 12 bytes on stack + 2 bytes if a BSR/bsr calls it
;*********************************************************************
SpSub:       staa    FSTAT        ;[PwO] register command
             nop                  ;[O] wait min 4~ from w cycle to r
             nop                  ;[O]
             nop                  ;[O]
             brclr   FSTAT,CCIF,* ;[rfPPP] wait for queued commands to finish
SpSubEnd:    rts                  ;back into DoOnStack in flash

;*********************************************************************
;* User Pseudo-vector Equates (just before protected block)
;*  real vectors point here, each pseudo-vector is a bra instruction
;*  to the user's ISR.
;*********************************************************************
BSRTable:
uINT80:      bsr    ISRHandler    ; INT80 reserved
uINT82:      bsr    ISRHandler    ; INT82 reserved
uINT84:      bsr    ISRHandler    ; INT84 reserved
uINT86:      bsr    ISRHandler    ; INT86 reserved
uINT88:      bsr    ISRHandler    ; INT88 reserved
uINT8x:      bsr    ISRHandler    ; INT8A reserved
uPwm_sd:     bsr    ISRHandler    ; PWM Shutdown
uPortP:      bsr    ISRHandler    ; Port P Interrupt
uCan4_Tx:    bsr    ISRHandler    ; MSCAN 4 transmit
uCan4_Rx:    bsr    ISRHandler    ; MSCAN 4 receive
uCan4_err:   bsr    ISRHandler    ; MSCAN 4 errors
uCan4_Wup:   bsr    ISRHandler    ; MSCAN 4 wake-up
uCan3_Tx:    bsr    ISRHandler    ; MSCAN 3 transmit
uCan3_Rx:    bsr    ISRHandler    ; MSCAN 3 receive
uCan3_err:   bsr    ISRHandler    ; MSCAN 3 errors
uCan3_Wup:   bsr    ISRHandler    ; MSCAN 3 wake-up
uCan2_Tx:    bsr    ISRHandler    ; MSCAN 2 transmit
uCan2_Rx:    bsr    ISRHandler    ; MSCAN 2 receive
uCan2_err:   bsr    ISRHandler    ; MSCAN 2 errors
uCan2_Wup:   bsr    ISRHandler    ; MSCAN 2 wake-up
uCan1_Tx:    bsr    ISRHandler    ; MSCAN 1 transmit
uCan1_Rx:    bsr    ISRHandler    ; MSCAN 1 receive
uCan1_err:   bsr    ISRHandler    ; MSCAN 1 errors
uCan1_Wup:   bsr    ISRHandler    ; MSCAN 1 wake-up
uCan0_Tx:    bsr    ISRHandler    ; MSCAN 0 transmit
uCan0_Rx:    bsr    ISRHandler    ; MSCAN 0 receive
uCan0_err:   bsr    ISRHandler    ; MSCAN 0 errors
uCan0_Wup:   bsr    ISRHandler    ; MSCAN 0 wake-up
uFlash:      bsr    ISRHandler    ; Flash
uEeprom:     bsr    ISRHandler    ; EEPROM
uSpi2:       bsr    ISRHandler    ; SPI2
uSpi1:       bsr    ISRHandler    ; SPI1
uIic:        bsr    ISRHandler    ; IIC Bus
uBdlc:       bsr    ISRHandler    ; BDLC
uScme:       bsr    ISRHandler    ; SCME
uCrg:        bsr    ISRHandler    ; CRG lock
uPulseB_of:  bsr    ISRHandler    ; Pulse acc overf
uDwnCnt_un:  bsr    ISRHandler    ; Mod Dwn Count undrflw
uPortH:      bsr    ISRHandler    ; Port H
uPortJ:      bsr    ISRHandler    ; port J
uAdc1:       bsr    ISRHandler    ; ADC1
uAdc0:       bsr    ISRHandler    ; ADC0
uSci1:       bsr    ISRHandler    ; SCI1
;             bsr    ISRHandler    ; SCI0, used for monitor
uSpi0:       bsr    ISRHandler    ; SPI0
uPulseA_in:  bsr    ISRHandler    ; Pulse acc input
uPulseA_of:  bsr    ISRHandler    ; Pulse acc overf
uTimer:      bsr    ISRHandler    ; Timer overflow
uTim_ch_7:   bsr    ISRHandler    ; Timer channel 7
uTim_ch_6:   bsr    ISRHandler    ; Timer channel 6
uTim_ch_5:   bsr    ISRHandler    ; Timer channel 5
uTim_ch_4:   bsr    ISRHandler    ; Timer channel 4
uTim_ch_3:   bsr    ISRHandler    ; Timer channel 3
uTim_ch_2:   bsr    ISRHandler    ; Timer channel 2
uTim_ch_1:   bsr    ISRHandler    ; Timer channel 1
uTim_ch_0:   bsr    ISRHandler    ; Timer channel 0
uRti:        bsr    ISRHandler    ; Real time
uIrq:        bsr    ISRHandler    ; IRQ
uXirq:       bsr    ISRHandler    ; XIRQ
;             bsr    ISRHandler    ; SWI, used for breakpoints
uIllegal:    bsr    ISRHandler    ; illegal
uCop:        bsr    ISRHandler    ; cop fail
uClkFail:    bsr    ISRHandler    ; clock fail
uReset:      bsr    ISRHandler    ; RESET

;*********************************************************************
;* User Pseudo-vector Equates (just before protected block)
;*  real vectors point here, each pseudo-vector is a bra instruction
;*  to the user's ISR.
;*********************************************************************
ISRHandler:  pulx
             ldy     (PVecTable-BSRTable-2),X
             cpy     #$FFFF
             beq     BadVector
             jmp     ,Y

;*********************************************************************
;* Invalid (erased) vector fetched
;*   low byte of vector address is Status (passed in A to ReenterMon)
;*********************************************************************
BadVector:   leax   (PVecTable-BSRTable-2),X
             xgdx
             tba                  ;get low byte of vector address in A
             jmp    ReenterMon    ;and enter monitor
EndMonitor:

  IF EndMonitor >= FProtStart
    FAIL "Monitor code overflows into protection/security area"
  ENDIF
;*********************************************************************
;* FLASH configuration: protection, security
;*********************************************************************
             org    FProtStart   ; enable protection
             fdb    $FFFF        ; Skip Backdoor Key
             fdb    $FFFF        ; Skip Backdoor Key
             fdb    $FFFF        ; Skip Backdoor Key
             fdb    $FFFF        ; Skip Backdoor Key

             fdb    $FFFF        ; Skip Reserved

             fcb    $FF          ; protection block 3
             fcb    $FF          ; protection block 2
             fcb    $FF          ; protection block 1
             fcb    FProtBlksz   ; protection block 0
             fcb    $FF          ; Skip
             fcb    FSecure      ; set security and backdoor access
;*********************************************************************
;* Jump table for external use of routines.
;*********************************************************************
             org    VectorTable-22 ;immediately before the vector table
             jsr    PutChar
             jsr    GetChar
             jsr    EraseAllCmd
             jsr    DoOnStack
             jsr    WriteD2IX

             org    VectorTable-4
             fdb    softwareID1
             fdb    softwareID2
;
;*********************************************************************
;* Define all vectors even if program doesn't use them all
;*********************************************************************
; ??? change all references to uINIT80 type not part specific.

              org    VectorTable
vecINT80:     fdb    uINT8x          ; INT80 reserved         */
vecINT82:     fdb    uINT8x          ; INT82 reserved         */
vecINT84:     fdb    uINT8x          ; INT84 reserved         */
vecINT86:     fdb    uINT8x          ; INT86 reserved         */
vecINT88:     fdb    uINT8x          ; INT88 reserved         */
vecINT8A:     fdb    uINT8x          ; INT8A reserved         */
vecPwm_sd:    fdb    uPwm_sd         ; PWM Shutdown           */
vecPortP:     fdb    uPortP          ; Port P Interrupt       */
vecCan4_Tx:   fdb    uCan4_Tx        ; MSCAN 4 transmit       */
vecCan4_Rx:   fdb    uCan4_Rx        ; MSCAN 4 receive        */
vecCan4_err:  fdb    uCan4_err       ; MSCAN 4 errors         */
vecCan4_Wup:  fdb    uCan4_Wup       ; MSCAN 4 wake-up        */
vecCan3_Tx:   fdb    uCan3_Tx        ; MSCAN 3 transmit       */
vecCan3_Rx:   fdb    uCan3_Rx        ; MSCAN 3 receive        */
vecCan3_err:  fdb    uCan3_err       ; MSCAN 3 errors         */
vecCan3_Wup:  fdb    uCan3_Wup       ; MSCAN 3 wake-up        */
vecCan2_Tx:   fdb    uCan2_Tx        ; MSCAN 2 transmit       */
vecCan2_Rx:   fdb    uCan2_Rx        ; MSCAN 2 receive        */
vecCan2_err:  fdb    uCan2_err       ; MSCAN 2 errors         */
vecCan2_Wup:  fdb    uCan2_Wup       ; MSCAN 2 wake-up        */
vecCan1_Tx:   fdb    uCan1_Tx        ; MSCAN 1 transmit       */
vecCan1_Rx:   fdb    uCan1_Rx        ; MSCAN 1 receive        */
vecCan1_err:  fdb    uCan1_err       ; MSCAN 1 errors         */
vecCan1_Wup:  fdb    uCan1_Wup       ; MSCAN 1 wake-up        */
vecCan0_Tx:   fdb    uCan0_Tx        ; MSCAN 0 transmit       */
vecCan0_Rx:   fdb    uCan0_Rx        ; MSCAN 0 receive        */
vecCan0_err:  fdb    uCan0_err       ; MSCAN 0 errors         */
vecCan0_Wup:  fdb    uCan0_Wup       ; MSCAN 0 wake-up        */
vecFlash:     fdb    uFlash          ; Flash                  */
vecEeprom:    fdb    uEeprom         ; EEPROM                 */
vevSpi2:      fdb    uSpi2           ; SPI2                  */
vevSpi1:      fdb    uSpi1           ; SPI1                  */
vecIic:       fdb    uIic            ; IIC Bus               */
vecBdlc:      fdb    uBdlc           ; BDLC                  */
vecScme:      fdb    uScme           ; SCME                  */
vecCrg:       fdb    uCrg            ; CRG lock              */
vecPulseB_of: fdb    uPulseB_of      ; Pulse acc overf       */
vecDwnCnt_un: fdb    uDwnCnt_un      ; Mod Dwn Count undrflw */
vecPortH:     fdb    uPortH          ; Port H                */
vecPortJ:     fdb    uPortJ          ; port J                */
vecAdc1:      fdb    uAdc1           ; ADC1                  */
vecAdc0:      fdb    uAdc0           ; ADC0                  */
vecSci1:      fdb    uSci1           ; SCI1                  */
vecSci0:      fdb    uSci0           ; SCI0                  */
vevSpi0:      fdb    uSpi0           ; SPI0                  */
vecPulseA_in: fdb    uPulseA_in      ; Pulse acc input       */
vecPulseA_of: fdb    uPulseA_of      ; Pulse acc overf       */
vecTimer:     fdb    uTimer          ; Timer overflow        */
vecTim_ch_7:  fdb    uTim_ch_7       ; Timer channel 7       */
vecTim_ch_6:  fdb    uTim_ch_6       ; Timer channel 6       */
vecTim_ch_5:  fdb    uTim_ch_5       ; Timer channel 5       */
vecTim_ch_4:  fdb    uTim_ch_4       ; Timer channel 4       */
vecTim_ch_3:  fdb    uTim_ch_3       ; Timer channel 3       */
vecTim_ch_2:  fdb    uTim_ch_2       ; Timer channel 2       */
vecTim_ch_1:  fdb    uTim_ch_1       ; Timer channel 1       */
vecTim_ch_0:  fdb    uTim_ch_0       ; Timer channel 0       */
vecRti:       fdb    uRti            ; Real time             */
vecIrq:       fdb    uIrq            ; IRQ                   */
vecXirq:      fdb    uXirq           ; XIRQ                  */
vecSwi:       fdb    Breakpoint      ; SWI                   */
vecIllegal:   fdb    uIllegal        ; illegal               */
vecCop:       fdb    coldStart       ; cop fail  (uCop)      */
vecClkFail:   fdb    uClkFail        ; clock fail            */
vecReset:     fdb    coldStart       ; RESET                 */

;.nolist                           ;to skip symbol table

;*****************************************************************
